<el-input
	type="textarea"
	:placeholder="data.{{{widget.widget_id}}}.placeholder"
	:rows="data.{{{widget.widget_id}}}.rows"
	:autosize="data.{{{widget.widget_id}}}.autosize"
	v-model="state.{{{widget.widget_id}}}.value"
>
</el-input>
