from supervisely.nn.inference.gui.gui import BaseInferenceGUI, InferenceGUI
from supervisely.nn.inference.gui.serving_gui import ServingGUI
from supervisely.nn.inference.gui.serving_gui_template import ServingGUITemplate
